@@ -1,10 +1,16 @@
-use crate::{entry, Entry, EntryRef};
-use std::borrow::Cow;
+use std::{
+    borrow::Cow,
+    path::{Path, PathBuf},
+};
 
-use crate::entry::PathspecMatch;
-use crate::walk::{Context, Error, ForDeletionMode, Options};
 use bstr::{BStr, BString, ByteSlice};
-use std::path::{Path, PathBuf};
+
+use crate::{
+    entry,
+    entry::PathspecMatch,
+    walk::{Context, Error, ForDeletionMode, Options},
+    Entry, EntryRef,
+};
 
 /// Classify the `worktree_relative_root` path and return the first `PathKind` that indicates that
 /// it isn't a directory, leaving `buf` with the path matching the returned `PathKind`,
@@ -13,24 +19,18 @@ pub fn root(
     worktree_root: &Path,
     buf: &mut BString,
     worktree_relative_root: &Path,
-    options: Options,
+    options: Options<'_>,
     ctx: &mut Context<'_>,
 ) -> Result<(Outcome, bool), Error> {
     buf.clear();
     let mut last_length = None;
     let mut path_buf = worktree_root.to_owned();
     // These initial values kick in if worktree_relative_root.is_empty();
-    let file_kind = path_buf.symlink_metadata().map(|m| m.file_type().into()).ok();
-    let pathspec_orig = std::mem::replace(
-        ctx.pathspec,
-        gix_pathspec::Search::from_specs(None, None, "".as_ref()).expect("empty is valid"),
-    );
-    let res = path(&mut path_buf, buf, 0, file_kind, || None, options, ctx);
-    *ctx.pathspec = pathspec_orig;
-    let mut out = res?;
+    let file_kind = path_buf.symlink_metadata().ok().map(|m| m.file_type().into());
+    let mut out = path(&mut path_buf, buf, 0, file_kind, || None, options, ctx)?;
     let worktree_root_is_repository = out
         .disk_kind
-        .map_or(false, |kind| matches!(kind, entry::Kind::Repository));
+        .is_some_and(|kind| matches!(kind, entry::Kind::Repository));
 
     for component in worktree_relative_root.components() {
         if last_length.is_some() {
@@ -38,7 +38,7 @@ pub fn root(
         }
         path_buf.push(component);
         buf.extend_from_slice(gix_path::os_str_into_bstr(component.as_os_str()).expect("no illformed UTF8"));
-        let file_kind = path_buf.symlink_metadata().map(|m| m.file_type().into()).ok();
+        let file_kind = path_buf.symlink_metadata().ok().map(|m| m.file_type().into());
 
         out = path(
             &mut path_buf,
@@ -59,9 +59,6 @@ pub fn root(
         }
         last_length = Some(buf.len());
     }
-    if out.pathspec_match.is_none() {
-        out.pathspec_match = Some(PathspecMatch::Always);
-    }
     Ok((out, worktree_root_is_repository))
 }
 /// The product of [`path()`] calls.
@@ -131,6 +128,8 @@ impl<'a> EntryRef<'a> {
 ///
 /// Returns `(status, file_kind, pathspec_matches_how)` to identify the `status` on disk, along with a classification `file_kind`,
 /// and if `file_kind` is not a directory, the way the pathspec matched with `pathspec_matches_how`.
+///
+/// Note that non-files are pruned by default.
 pub fn path(
     path: &mut PathBuf,
     rela_path: &mut BString,
@@ -144,8 +143,9 @@ pub fn path(
         for_deletion,
         classify_untracked_bare_repositories,
         symlinks_to_directories_are_ignored_like_directories,
+        worktree_relative_worktree_dirs,
         ..
-    }: Options,
+    }: Options<'_>,
     ctx: &mut Context<'_>,
 ) -> Result<Outcome, Error> {
     let mut out = Outcome {
@@ -170,10 +170,15 @@ pub fn path(
                 .as_mut()
                 .map_or(Ok(None), |stack| {
                     stack
-                        .at_entry(rela_path.as_bstr(), disk_kind.map(|ft| ft.is_dir()), ctx.objects)
+                        .at_entry(
+                            rela_path.as_bstr(),
+                            disk_kind.map(|ft| is_dir_to_mode(ft.is_dir())),
+                            ctx.objects,
+                        )
                         .map(|platform| platform.excluded_kind())
                 })
                 .map_err(Error::ExcludesAccess)?
+                .filter(|_| filename_start_idx > 0)
             {
                 out.status = entry::Status::Ignored(excluded);
             }
@@ -181,10 +186,9 @@ pub fn path(
         out.property = entry::Property::DotGit.into();
         return Ok(out);
     }
-    let pathspec_could_match = rela_path.is_empty()
-        || ctx
-            .pathspec
-            .can_match_relative_path(rela_path.as_bstr(), disk_kind.map(|ft| ft.is_dir()));
+    let pathspec_could_match = ctx
+        .pathspec
+        .can_match_relative_path(rela_path.as_bstr(), disk_kind.map(|ft| ft.is_dir()));
     if !pathspec_could_match {
         return Ok(out.with_status(entry::Status::Pruned));
     }
@@ -196,6 +200,18 @@ pub fn path(
     );
     let mut kind = uptodate_index_kind.or(disk_kind).or_else(on_demand_disk_kind);
 
+    // We always check the pathspec to have the value filled in reliably.
+    out.pathspec_match = ctx
+        .pathspec
+        .pattern_matching_relative_path(rela_path.as_bstr(), kind.map(|ft| ft.is_dir()), ctx.pathspec_attributes)
+        .map(Into::into);
+
+    if worktree_relative_worktree_dirs.is_some_and(|worktrees| worktrees.contains(&*rela_path)) {
+        return Ok(out
+            .with_kind(Some(entry::Kind::Repository), None)
+            .with_status(entry::Status::Tracked));
+    }
+
     let maybe_status = if property.is_none() {
         (index_kind.map(|k| k.is_dir()) == kind.map(|k| k.is_dir())).then_some(entry::Status::Tracked)
     } else {
@@ -203,53 +219,24 @@ pub fn path(
         Some(entry::Status::Pruned)
     };
 
-    // We always check the pathspec to have the value filled in reliably.
-    out.pathspec_match = ctx
-        .pathspec
-        .pattern_matching_relative_path(rela_path.as_bstr(), kind.map(|ft| ft.is_dir()), ctx.pathspec_attributes)
-        .map(Into::into);
-
     let is_dir = if symlinks_to_directories_are_ignored_like_directories
         && ctx.excludes.is_some()
-        && kind.map_or(false, |ft| ft == entry::Kind::Symlink)
+        && kind == Some(entry::Kind::Symlink)
     {
-        path.metadata().ok().map(|md| md.is_dir()).or(Some(false))
+        path.metadata().ok().map(|md| is_dir_to_mode(md.is_dir()))
     } else {
-        kind.map(|ft| ft.is_dir())
+        kind.map(|ft| is_dir_to_mode(ft.is_dir()))
     };
 
     let mut maybe_upgrade_to_repository = |current_kind, find_harder: bool| {
-        if recurse_repositories {
-            return current_kind;
-        }
-        if find_harder {
-            let mut is_nested_repo = gix_discover::is_git(path).is_ok();
-            if is_nested_repo {
-                let git_dir_is_our_own =
-                    gix_path::realpath_opts(path, ctx.current_dir, gix_path::realpath::MAX_SYMLINKS)
-                        .ok()
-                        .map_or(false, |realpath_candidate| realpath_candidate == ctx.git_dir_realpath);
-                is_nested_repo = !git_dir_is_our_own;
-            }
-            if is_nested_repo {
-                return Some(entry::Kind::Repository);
-            }
-        }
-        path.push(gix_discover::DOT_GIT_DIR);
-        let mut is_nested_nonbare_repo = gix_discover::is_git(path).is_ok();
-        if is_nested_nonbare_repo {
-            let git_dir_is_our_own = gix_path::realpath_opts(path, ctx.current_dir, gix_path::realpath::MAX_SYMLINKS)
-                .ok()
-                .map_or(false, |realpath_candidate| realpath_candidate == ctx.git_dir_realpath);
-            is_nested_nonbare_repo = !git_dir_is_our_own;
-        }
-        path.pop();
-
-        if is_nested_nonbare_repo {
-            Some(entry::Kind::Repository)
-        } else {
-            current_kind
-        }
+        maybe_upgrade_to_repository(
+            current_kind,
+            find_harder,
+            recurse_repositories,
+            path,
+            ctx.current_dir,
+            ctx.git_dir_realpath,
+        )
     };
     if let Some(status) = maybe_status {
         if kind == Some(entry::Kind::Directory) && index_kind == Some(entry::Kind::Repository) {
@@ -262,6 +249,7 @@ pub fn path(
     if let Some(excluded) = ctx
         .excludes
         .as_mut()
+        .filter(|_| !rela_path.is_empty())
         .map_or(Ok(None), |stack| {
             stack
                 .at_entry(rela_path.as_bstr(), is_dir, ctx.objects)
@@ -285,8 +273,17 @@ pub fn path(
                     ),
                 );
             }
-            if kind.map_or(false, |d| d.is_recursable_dir()) && out.pathspec_match.is_none() {
-                // we have patterns that didn't match at all, *yet*. We want to look inside.
+            if kind.is_some_and(|d| d.is_recursable_dir())
+                && (out.pathspec_match.is_none()
+                    || worktree_relative_worktree_dirs.is_some_and(|worktrees| {
+                        for_deletion.is_some()
+                            && worktrees
+                                .iter()
+                                .any(|dir| dir.starts_with_str(&*rela_path) && dir.get(rela_path.len()) == Some(&b'/'))
+                    }))
+            {
+                // We have patterns that didn't match at all, *yet*, or there are contained worktrees.
+                // We want to look inside.
                 out.pathspec_match = Some(PathspecMatch::Prefix);
             }
         }
@@ -298,7 +295,7 @@ pub fn path(
     debug_assert!(maybe_status.is_none());
     let mut status = entry::Status::Untracked;
 
-    if kind.map_or(false, |ft| ft.is_dir()) {
+    if kind.is_some_and(|ft| ft.is_dir()) {
         kind = maybe_upgrade_to_repository(kind, classify_untracked_bare_repositories);
     } else if out.pathspec_match.is_none() {
         status = entry::Status::Pruned;
@@ -306,6 +303,46 @@ pub fn path(
     Ok(out.with_status(status).with_kind(kind, index_kind))
 }
 
+pub fn maybe_upgrade_to_repository(
+    current_kind: Option<entry::Kind>,
+    find_harder: bool,
+    recurse_repositories: bool,
+    path: &mut PathBuf,
+    current_dir: &Path,
+    git_dir_realpath: &Path,
+) -> Option<entry::Kind> {
+    if recurse_repositories {
+        return current_kind;
+    }
+    if find_harder {
+        let mut is_nested_repo = gix_discover::is_git(path).is_ok();
+        if is_nested_repo {
+            let git_dir_is_our_own = gix_path::realpath_opts(path, current_dir, gix_path::realpath::MAX_SYMLINKS)
+                .ok()
+                .is_some_and(|realpath_candidate| realpath_candidate == git_dir_realpath);
+            is_nested_repo = !git_dir_is_our_own;
+        }
+        if is_nested_repo {
+            return Some(entry::Kind::Repository);
+        }
+    }
+    path.push(gix_discover::DOT_GIT_DIR);
+    let mut is_nested_nonbare_repo = gix_discover::is_git(path).is_ok();
+    if is_nested_nonbare_repo {
+        let git_dir_is_our_own = gix_path::realpath_opts(path, current_dir, gix_path::realpath::MAX_SYMLINKS)
+            .ok()
+            .is_some_and(|realpath_candidate| realpath_candidate == git_dir_realpath);
+        is_nested_nonbare_repo = !git_dir_is_our_own;
+    }
+    path.pop();
+
+    if is_nested_nonbare_repo {
+        Some(entry::Kind::Repository)
+    } else {
+        current_kind
+    }
+}
+
 /// Note that `rela_path` is used as buffer for convenience, but will be left as is when this function returns.
 /// Also note `maybe_file_type` will be `None` for entries that aren't up-to-date and files, for directories at least one entry must be uptodate.
 /// Returns `(maybe_file_type, Option<index_file_type>, flags)`, with the last option being a flag set only for sparse directories in the index.
@@ -356,7 +393,7 @@ fn resolve_file_type_with_index(
             }
             Some(entry) => {
                 let icase_dir = index.entry_closest_to_directory_icase(rela_path.as_bstr(), true, accelerate);
-                let directory_matches_exactly = icase_dir.map_or(false, |dir| {
+                let directory_matches_exactly = icase_dir.is_some_and(|dir| {
                     let path = dir.path(index);
                     let slash_idx = path.rfind_byte(b'/').expect("dir");
                     path[..slash_idx].as_bstr() == rela_path
@@ -399,7 +436,7 @@ fn resolve_file_type_with_index(
                 if all_excluded_from_worktree_non_cone
                     || one_index_signalling_with_cone
                         .filter(|_| kind.is_none())
-                        .map_or(false, |idx| index.entries()[idx].mode.is_sparse())
+                        .is_some_and(|idx| index.entries()[idx].mode.is_sparse())
                 {
                     special_property = Some(entry::Property::TrackedExcluded);
                 }
@@ -418,3 +455,11 @@ fn is_eq(lhs: &BStr, rhs: impl AsRef<BStr>, ignore_case: bool) -> bool {
         lhs == rhs.as_ref()
     }
 }
+
+fn is_dir_to_mode(is_dir: bool) -> gix_index::entry::Mode {
+    if is_dir {
+        gix_index::entry::Mode::DIR
+    } else {
+        gix_index::entry::Mode::FILE
+    }
+}
